// Project 50 - 24LC512 I2C EEPROM

#include <avr/io.h>
#include <util/delay.h>

void I2Cenable()
// enable I2C bus
{
    TWBR = 72; // 100 kHz I2C bus
    TWCR |= (1 << TWEN); // enable I2C on PORTC4 and 5
}

void I2Cwait()
// wait until I2C finishes an operation
{
    // wait until bit TWINT in TWCR is set to 1
    while (!(TWCR & (1<<TWINT))); 
}

void I2CstartWait(unsigned char address)
{
    uint8_t status;
    while (1)
    {
        // send START condition
        TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN);
    
        // wait until transmission completed
    	I2Cwait();
    
    	// check value of TWSR, and mask out status bits
    	status = TWSR & 0b11111000;
    	if ((status != 0b00001000) && (status != 0b00010000)) continue;
    
    	// send device address
    	TWDR = address;
    	TWCR = (1<<TWINT) | (1<<TWEN);
    
    	// wait until transmission completed
    	I2Cwait();
    
    	// check value of TWSR, and mask out status bits
    	status = TWSR & 0b11111000;
    	if ((status == 0b00100000 )||(status == 0b01011000)) 
    	{    	    
            // slave device is busy, so send stop condition to terminate write operation
            TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWSTO);
	        
            // wait until stop condition is executed and I2C bus released
            while(TWCR & (1<<TWSTO));
            continue;
        }
        break;
    }
}

void I2Cstop()
// stop I2C bus and release GPIO pins
{
    // clear interrupt, enable I2C, generate stop condition
    TWCR |= (1 << TWINT)|(1 << TWEN)|(1 << TWSTO);
}

void I2Cwrite(uint8_t data)
// send 'data' to I2C bus
{
    TWDR = data;
    TWCR |= (1 << TWINT)|(1 << TWEN);
    I2Cwait();
}

uint8_t I2Cread()
// read incoming byte of data from I2C bus
{
    TWCR |= (1 << TWINT)|(1 << TWEN);
    I2Cwait();
    // incoming byte is placed in TWDR register
    return TWDR;	
}

uint8_t I2CreadACK()
// read incoming byte of data from I2C bus and ACK signal
{
    TWCR |= (1 << TWINT)|(1 << TWEN)|(1 << TWEA);
    I2Cwait();
    // incoming byte is placed in TWDR register
    return TWDR;	
}

void writeMAX7219(char hexdata1, char hexdata2)
// sends two bytes in hexadecimal to MAX7219
{
    PORTB &= ~(1 << PORTB0);    // latch pin low
    SPDR = hexdata1;            // send value of hexdata1
    while(!(SPSR & (1<<SPIF))); // wait for SPI transmission to finish
    SPDR = hexdata2;            // send value of hexdata2
    while(!(SPSR & (1<<SPIF))); // wait for SPI transmission to finish
    PORTB |= (1 << PORTB0);     // latch pin high
}

void blankMAX7219()
// blanks all digits
{
    uint8_t i;
    for (i=1; i<9; i++)	    // blank all digits
    {
        writeMAX7219(i,15);
    }
}

void initMAX7219()
// setup MAX7219 for use
{
    PORTB |= (1 << 0); 		// latch pin high (SS)
    SPCR = (1<<SPE)|(1<<MSTR);  		// set up SPI bus for MAX7219
    writeMAX7219(0x09,0xFF); // mode decode for digits (table 4 page 7 - 0x09 + 0xFF)
    writeMAX7219(0x0B,0x07); // set scan limit to 8 digits - 0x09 + 0xFF)
    writeMAX7219(0x0A,0x01); // set intensity to 8 - 0x0A + 0x08)
    writeMAX7219(0x0C,0x01); // mode display on (table 4 page 7 - 0x09 + 0xFF)
    blankMAX7219();
} 

void dispMAX7219(uint8_t digit, uint8_t number, uint8_t dp)
// displays "number" in location "digit" with decimal point on/off
// digit: 1~8 for location 1~8
// number: 0~15 for 0~9, - E, H, L, P, blank
// dp: 1 on, 0 off
{
    if (dp==1)      // add decimal point
    { 
        number = number + 128; 
    } 
    writeMAX7219(digit, number);
}

void numberMAX7219(uint32_t value)    
// Displays a number between 0-99999999 on MAX7219-controlled 8 digit display
{
    uint8_t digits[9];
    uint8_t i = 1;
	
    for (i=1; i<9; i++)
    {
        digits[i]=15;
    }
	
    i = 1;
    while (value > 0) 			 // continue until value > 0
    {
        digits[i] = value % 10;  // determine and store last digit of number
        value = value / 10;      // divide value by 10
	i++;
    }
    for (i=1; i<9; i++)
    {
        dispMAX7219(i, digits[i],0);
    }
}

int main(void)
{
    uint16_t i;
    uint16_t j;
	
    DDRB = 0b11111111; // set PORTB as outputs	
    I2Cenable();
    initMAX7219();	
    while (1) 
    {
        dispMAX7219(0,10,0);
        for (i = 0; i<256; i++) 
        { 
            I2CstartWait(0x50 << 1); // 0x50 << 1 - 0b10100000
            I2Cwrite(i >> 8);
            I2Cwrite(i);
            I2Cwrite(i);			
            I2Cstop();
            numberMAX7219(i);
            _delay_ms(1);			
        }

        // now retrieve them and display on MAX7219

        for (i = 0; i<256; i++) 
        {
            I2CstartWait(0xA0); // write address
            I2Cwrite(i >> 8);
            I2Cwrite(i);
            I2CstartWait((0x50<<1)+1); // read address - 0b10100001
            j = I2Cread();
            I2Cstop();
            numberMAX7219(j);
            _delay_ms(5);
        }
        _delay_ms(50);
    }
}