// Project 59 - Analog Clock with Servos

#include <avr/io.h>
#include <util/delay.h>

// variabes to store time and date
uint8_t hours, minutes, seconds, dow, dom, mo, years;	

void I2Cenable()
// enable I2C bus
{
    TWBR = 72; // 100 kHz I2C bus
    TWCR |= (1 << TWEN); // enable I2C on PORTC4 and 5
}

void I2Cwait()
// wait until I2C finishes an operation
{
    // wait until bit TWINT in TWCR is set to 1
    while (!(TWCR & (1<<TWINT))); 
}

void I2CstartWait(unsigned char address)
{
    uint8_t status;
    while (1)
    {
        // send START condition
        TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN);
    
    	// wait until transmission completed
    	I2Cwait();
    
    	// check value of TWSR, and mask out status bits
    	status = TWSR & 0b11111000;
    	if ((status != 0b00001000) && (status != 0b00010000)) continue;
    
    	// send device address
    	TWDR = address;
    	TWCR = (1<<TWINT) | (1<<TWEN);
    
    	// wait until transmission completed
    	I2Cwait();
    
    	// check value of TWSR, and mask out status bits
    	status = TWSR & 0b11111000;
    	if ((status == 0b00100000 )||(status == 0b01011000)) 
    	{    	    
            // secondary device is busy, so send stop condition to terminate write operation
            TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWSTO);
            // wait until stop condition is executed and I2C bus released
            while(TWCR & (1<<TWSTO));
            continue;
        }
        break;
    }
}

void I2Cstop()
// stop I2C bus and release GPIO pins
{
    // clear interrupt, enable I2C, generate stop condition
    TWCR |= (1 << TWINT)|(1 << TWEN)|(1 << TWSTO);
}

void I2Cwrite(uint8_t data)
// send 'data' to I2C bus
{
    TWDR = data;
    TWCR |= (1 << TWINT)|(1 << TWEN);
    I2Cwait();
}

uint8_t I2Cread()
// read incoming byte of data from I2C bus
{
    TWCR |= (1 << TWINT)|(1 << TWEN);
    I2Cwait();
    // incoming byte is placed in TWDR register
    return TWDR;	
}

uint8_t I2CreadACK()
// read incoming byte of data from I2C bus and ACK signal
{
    TWCR |= (1 << TWINT)|(1 << TWEN)|(1 << TWEA);
    I2Cwait();
    // incoming byte is placed in TWDR register
    return TWDR;	
}

uint8_t decimalToBcd(uint8_t val)
// Convert integer to BCD
{
    return((val/10*16)+(val%10));
}

uint8_t bcdToDec(uint8_t val)
// Convert BCD to integer
{
    return((val/16*10)+(val%16));
}

void setTimeDS3231(uint8_t hh, uint8_t mm, uint8_t ss, uint8_t dw, uint8_t dd, uint8_t mo, uint8_t yy)
// set the time on DS3231
{
    I2CstartWait(0xD0); // DS3231 write
    I2Cwrite(0x00);		// start with hours register
    I2Cwrite(decimalToBcd(ss));  // seconds
    I2Cwrite(decimalToBcd(mm));  // minutes
    I2Cwrite(decimalToBcd(hh));  // hours
    I2Cwrite(decimalToBcd(dw));  // day of week
    I2Cwrite(decimalToBcd(dd));  // date
    I2Cwrite(decimalToBcd(mo));	 // month
    I2Cwrite(decimalToBcd(yy));	 // year
    I2Cstop();	
}

void readTimeDS3231()
// retrieve time and date from DS3231
{
    I2CstartWait(0xD0); // DS3231 write
    I2Cwrite(0x00);		// seconds register
    I2CstartWait(0xD1); // DS3231 read
    seconds = bcdToDec(I2CreadACK());	
    minutes = bcdToDec(I2CreadACK());	
    hours = bcdToDec(I2CreadACK());	
    dow = bcdToDec(I2CreadACK());	
    dom = bcdToDec(I2CreadACK());
    mo = bcdToDec(I2CreadACK());
    years = bcdToDec(I2CreadACK());
}

void initPWM()
{    // activate PWM
    TCCR1A |= (1 << WGM11);
    TCCR1B |= (1 << WGM12)|(1 << WGM13)|(1 << CS11); 
	
    // connect PWM to PB1 and PB2
    TCCR1A |= (1 << COM1A1)|(1 << COM1B1); // PWM to OCR1A - PB1 and OCR1B - PB2
    ICR1=39999; 
}

void servoAngleA(uint8_t angle)
{    // rotate servo on OCR1A to 'angle' position
    OCR1A = ((angle-239.95)/-0.05); // convert angle to OCR1A (duty cycle) value
}

void servoAngleB(uint8_t angle)
{    // rotate servo on OCR1B to 'angle' position
    OCR1B = ((angle-239.95)/-0.05); // convert angle to OCR1A (duty cycle) value
}

void displayServoTime()
{    // displays hours on servo A, minutes on servo B
    uint8_t _hours;
    uint8_t _minutes;
    
    _hours = hours * 15; // each hour = 15 servo degrees
    servoAngleA(_hours);
    _minutes = minutes * 3; // each minute = 3 servo degrees
    servoAngleB(_minutes);
}

int main()
{
    DDRB = 0b11111111; // set PORTB as outputs	
    I2Cenable();
    initPWM();
	
    // uncomment to set time and date, then comment and re-flash code
    // setTimeDS3231(9,13,0,5,29,4,21); // h,m,s,dow,dom,m,y
	
    while(1)
    {
        readTimeDS3231();
        displayServoTime();
        _delay_ms(1000);
    }
} 